/*
Section:  5) BCNF to 4NF
Script:   11) Getting Data Into 4NF
Basics:   We have our data set in BCNF.  Now to move to the next step.
Notes:    
*/

/*
Once we're in BCNF, going to 4NF means checking for multiple, independent,
multi-valued facts about an entity.  Member, Instrument, Label, AlbumXMember,
and Song are definitely out:  they don't have multiple facts.

Quick note:  SongNameHasNumeral is out of the data model again; we don't need it anymore.

We tend to focus on 2 cases:
1)  Ternary keys:  AlbumXMemberXInstrument, SongXMember
2)  Independent facts (none on our data model)

NEW structure:

Member
	- Name (PK)

Instrument
	- Name (PK)

Label
	- Name (PK)
	- Headquarters

Album
	- Year
	- ProjeKctName (NULL)
	- Type { Studio, Live }
	- AlbumName (PK)
	- LabelName (NULL)

AlbumXMember
	- AlbumName (PK)
	- MemberName (PK)
	- IsSessionOnly

AlbumXMemberXInstrument
	- AlbumName (PK)
	- MemberName (PK)
	- InstrumentName (PK)

Song
	- AlbumName (PK, AK2)
	- TrackNumber (AK2)
	- SongName (PK)

SongXMember
	- AlbumName (PK)
	- SongName (PK)
	- MemberName (PK)

TODO:
1)  Determine whether all entities are in 4NF.  Fix those which are not.
2)  Suppose we had the following entity, modeled in minimal format.
	Assume that instrument and projeKCt are fully independent:  instead of
	listing which instruments a person played for a particular projeKCt,
	we are listing the instruments the person knows how to play.
	How could we put this into 4NF?

*--------------------------------------------*
|   Member   |   Instrument   |   ProjeKCt   |
----------------------------------------------
| Fripp      | Guitar         | ProjeKCt 1   |
| Fripp      |                | ProjeKCt 2   | 
| Bruford    | Drums          | ProjeKCt 1   |
| Gunn       | Touch guitar   | ProjeKCt 1   |
| Gunn       | Talker         | ProjeKCt 2   |
| Gunn       | Guitar synth   |              |
*--------------------------------------------*

*/

--Sample code to test AlbumXMemberXInstrument, stripping out unnecessary attributes
IF NOT EXISTS
(
	SELECT 1
	FROM sys.tables
	WHERE
		name = 'Member'
		AND OBJECT_SCHEMA_NAME(object_id) = '4NF'
)
BEGIN
	CREATE TABLE [4NF].[Member]
	(
		[Name] VARCHAR(40) NOT NULL
	);

	ALTER TABLE [4NF].[Member] ADD CONSTRAINT [PK_4NF_Member] PRIMARY KEY CLUSTERED(Name);
END
GO

IF NOT EXISTS
(
	SELECT 1
	FROM sys.tables
	WHERE
		name = 'Album'
		AND OBJECT_SCHEMA_NAME(object_id) = '4NF'
)
BEGIN
	CREATE TABLE [4NF].[Album]
	(
		[Name] VARCHAR(120) NOT NULL
	);

	ALTER TABLE [4NF].[Album] ADD CONSTRAINT [PK_4NF_Album] PRIMARY KEY CLUSTERED(Name);
END
GO

IF NOT EXISTS
(
	SELECT 1
	FROM sys.tables
	WHERE
		name = 'Instrument'
		AND OBJECT_SCHEMA_NAME(object_id) = '4NF'
)
BEGIN
	CREATE TABLE [4NF].[Instrument]
	(
		[Name] VARCHAR(40) NOT NULL
	);

	ALTER TABLE [4NF].[Instrument] ADD CONSTRAINT [PK_4NF_Instrument] PRIMARY KEY CLUSTERED(Name);
END
GO

IF NOT EXISTS
(
	SELECT 1
	FROM sys.tables
	WHERE
		name = 'AlbumXMemberXInstrument'
		AND OBJECT_SCHEMA_NAME(object_id) = '4NF'
)
BEGIN
	CREATE TABLE [4NF].[AlbumXMemberXInstrument]
	(
		[AlbumName] VARCHAR(120) NOT NULL,
		[MemberName] VARCHAR(40) NOT NULL,
		[InstrumentName] VARCHAR(40) NOT NULL
	);

	ALTER TABLE [4NF].[AlbumXMemberXInstrument] ADD CONSTRAINT [PK_4NF_AlbumXMemberXInstrument] PRIMARY KEY CLUSTERED(AlbumName, MemberName, InstrumentName);
	ALTER TABLE [4NF].[AlbumXMemberXInstrument] ADD CONSTRAINT [FK_AlbumXMemberXInstrument_Album] FOREIGN KEY(AlbumName) REFERENCES [4NF].[Album](Name) ON UPDATE CASCADE;
	ALTER TABLE [4NF].[AlbumXMemberXInstrument] ADD CONSTRAINT [FK_AlbumXMemberXInstrument_Member] FOREIGN KEY(MemberName) REFERENCES [4NF].[Member](Name) ON UPDATE CASCADE;
	ALTER TABLE [4NF].[AlbumXMemberXInstrument] ADD CONSTRAINT [FK_AlbumXMemberXInstrument_Instrument] FOREIGN KEY(InstrumentName) REFERENCES [4NF].[Instrument](Name) ON UPDATE CASCADE;
END
GO

--Hypothesis:  we can split AlbumXMemberXInstrument into AlbumXMember
--and MemberXInstrument without any loss in recomposition.
IF NOT EXISTS
(
	SELECT 1
	FROM sys.tables
	WHERE
		name = 'AlbumXMember'
		AND OBJECT_SCHEMA_NAME(object_id) = '4NF'
)
BEGIN
	CREATE TABLE [4NF].[AlbumXMember]
	(
		[AlbumName] VARCHAR(120) NOT NULL,
		[MemberName] VARCHAR(40) NOT NULL
	);

	ALTER TABLE [4NF].[AlbumXMember] ADD CONSTRAINT [PK_4NF_AlbumXMember] PRIMARY KEY CLUSTERED(AlbumName, MemberName);
	ALTER TABLE [4NF].[AlbumXMember] ADD CONSTRAINT [FK_AlbumXMember_Album] FOREIGN KEY(AlbumName) REFERENCES [4NF].[Album](Name) ON UPDATE CASCADE;
	ALTER TABLE [4NF].[AlbumXMember] ADD CONSTRAINT [FK_AlbumXMember_Member] FOREIGN KEY(MemberName) REFERENCES [4NF].[Member](Name) ON UPDATE CASCADE;
END
GO

IF NOT EXISTS
(
	SELECT 1
	FROM sys.tables
	WHERE
		name = 'MemberXInstrument'
		AND OBJECT_SCHEMA_NAME(object_id) = '4NF'
)
BEGIN
	CREATE TABLE [4NF].[MemberXInstrument]
	(
		[MemberName] VARCHAR(40) NOT NULL,
		[InstrumentName] VARCHAR(40) NOT NULL
	);

	ALTER TABLE [4NF].[MemberXInstrument] ADD CONSTRAINT [PK_4NF_MemberXInstrument] PRIMARY KEY CLUSTERED(MemberName, InstrumentName);
	ALTER TABLE [4NF].[MemberXInstrument] ADD CONSTRAINT [FK_MemberXInstrument_Member] FOREIGN KEY(MemberName) REFERENCES [4NF].[Member](Name) ON UPDATE CASCADE;
	ALTER TABLE [4NF].[MemberXInstrument] ADD CONSTRAINT [FK_MemberXInstrument_Instrument] FOREIGN KEY(InstrumentName) REFERENCES [4NF].[Instrument](Name) ON UPDATE CASCADE;
END
GO


--Now let's insert some data and test our hypothesis
MERGE [4NF].[Instrument] AS tgt
USING
(
	SELECT
		Name
	FROM
	(
		VALUES
		('Bass'),
		('Devices'),
		('Drums'),
		('Flute'),
		('Guitar'),
		('Guitar synth'),
		('Keyboards'),
		('Mellotron'),
		('Percussion'),
		('Piano'),
		('Reeds'),
		('Sax'),
		('Stick'),
		('Synth'),
		('Talker'),
		('Touch guitar'),
		('Vocals'),
		('V-Drums'),
		('Vibes'),
		('Woodwind'),
		('Words'),
		('Words and Illumination')
		
	) v(Name)
) AS src
	ON tgt.Name = src.Name
WHEN NOT MATCHED BY TARGET THEN INSERT
(
	Name
)
VALUES
(
	src.Name
);

MERGE [4NF].[Member] AS tgt
USING
(
	SELECT
		Name
	FROM
	(
		VALUES
		('Tony Levin'),
		('Trey Gunn'),
		('Bill Bruford'),
		('Robert Fripp'),
		('Adrian Belew'),
		('Ian McDonald'),
		('Greg Lake'),
		('Michael Giles'),
		('Peter Sinfield'),
		('Peter Giles'),
		('Keith Tippett'),
		('Mel Collins'),
		('Gordon Haskell')
	) v(Name)
) AS src
	ON tgt.Name = src.Name
WHEN NOT MATCHED BY TARGET THEN INSERT
(
	Name
)
VALUES
(
	src.Name
);

MERGE [4NF].[Album] AS tgt
USING
(
	SELECT
		Name
	FROM
	(
		VALUES
		('Live at the Jazz Cafe'),
		('Live Groove'),
		('Space Groove'),
		('In the Court of the Crimson King'),
		('In the Wake of Poseidon')
	) v
	(
		Name
	)
) AS src
	ON tgt.Name = src.Name
WHEN NOT MATCHED BY TARGET THEN INSERT
(
	Name
)
VALUES
(
	src.Name
);

MERGE [4NF].[AlbumXMemberXInstrument] AS tgt
USING
(
	SELECT
		AlbumName,
		MemberName,
		InstrumentName
	FROM
	(
		VALUES
		('Live at the Jazz Cafe', 'Tony Levin', 'Bass'),
		('Live at the Jazz Cafe', 'Tony Levin', 'Stick'),
		('Live at the Jazz Cafe', 'Tony Levin', 'Synth'),
		('Live at the Jazz Cafe', 'Trey Gunn', 'Touch guitar'),
		('Live at the Jazz Cafe', 'Bill Bruford', 'Drums'),
		('Live at the Jazz Cafe', 'Robert Fripp', 'Guitar'),
		('Live Groove', 'Adrian Belew', 'V-Drums'),
		('Live Groove', 'Robert Fripp', 'Guitar'),
		('Live Groove', 'Trey Gunn', 'Touch guitar'),
		('Live Groove', 'Trey Gunn', 'Talker'),
		('Space Groove', 'Adrian Belew', 'V-Drums'),
		('Space Groove', 'Robert Fripp', 'Guitar'),
		('Space Groove', 'Trey Gunn', 'Touch guitar'),
		('Space Groove', 'Trey Gunn', 'Guitar synth'),
		('In the Court of the Crimson King', 'Robert Fripp', 'Guitar'),
		('In the Court of the Crimson King', 'Ian McDonald', 'Reeds'),
		('In the Court of the Crimson King', 'Ian McDonald', 'Woodwind'),
		('In the Court of the Crimson King', 'Ian McDonald', 'Vibes'),
		('In the Court of the Crimson King', 'Ian McDonald', 'Keyboards'),
		('In the Court of the Crimson King', 'Ian McDonald', 'Mellotron'),
		('In the Court of the Crimson King', 'Ian McDonald', 'Vocals'),
		('In the Court of the Crimson King', 'Greg Lake', 'Bass'),
		('In the Court of the Crimson King', 'Greg Lake', 'Vocals'),
		('In the Court of the Crimson King', 'Michael Giles', 'Drums'),
		('In the Court of the Crimson King', 'Michael Giles', 'Percussion'),
		('In the Court of the Crimson King', 'Michael Giles', 'Vocals'),
		('In the Court of the Crimson King', 'Peter Sinfield', 'Words and Illumination'),
		('In the Wake of Poseidon', 'Robert Fripp', 'Guitar'),
		('In the Wake of Poseidon', 'Robert Fripp', 'Mellotron'),
		('In the Wake of Poseidon', 'Robert Fripp', 'Devices'),
		('In the Wake of Poseidon', 'Greg lake', 'Vocals'),
		('In the Wake of Poseidon', 'Michael Giles', 'Drums'),
		('In the Wake of Poseidon', 'Peter Giles', 'Bass'),
		('In the Wake of Poseidon', 'Keith Tippett', 'Piano'),
		('In the Wake of Poseidon', 'Mel Collins', 'Sax'),
		('In the Wake of Poseidon', 'Mel Collins', 'Flute'),
		('In the Wake of Poseidon', 'Gordon Haskell', 'Vocals'),
		('In the Wake of Poseidon', 'Peter Sinfield', 'Words')
	) v(AlbumName, MemberName, InstrumentName)
) AS src
	ON tgt.AlbumName = src.AlbumName
	AND tgt.MemberName = src.MemberName
	AND tgt.InstrumentName = src.InstrumentName
WHEN NOT MATCHED BY TARGET THEN INSERT
(
	AlbumName,
	MemberName,
	InstrumentName
)
VALUES
(
	src.AlbumName,
	src.MemberName,
	src.InstrumentName
);

MERGE [4NF].[AlbumXMember] AS tgt
USING
(
	SELECT
		AlbumName,
		MemberName
	FROM
	(
		VALUES
		('Live at the Jazz Cafe', 'Tony Levin'),
		('Live at the Jazz Cafe', 'Trey Gunn'),
		('Live at the Jazz Cafe', 'Bill Bruford'),
		('Live at the Jazz Cafe', 'Robert Fripp'),
		('Live Groove', 'Adrian Belew'),
		('Live Groove', 'Robert Fripp'),
		('Live Groove', 'Trey Gunn'),
		('Space Groove', 'Adrian Belew'),
		('Space Groove', 'Robert Fripp'),
		('Space Groove', 'Trey Gunn'),
		('In the Court of the Crimson King', 'Robert Fripp'),
		('In the Court of the Crimson King', 'Ian McDonald'),
		('In the Court of the Crimson King', 'Greg Lake'),
		('In the Court of the Crimson King', 'Michael Giles'),
		('In the Court of the Crimson King', 'Peter Sinfield'),
		('In the Wake of Poseidon', 'Robert Fripp'),
		('In the Wake of Poseidon', 'Greg lake'),
		('In the Wake of Poseidon', 'Michael Giles'),
		('In the Wake of Poseidon', 'Peter Giles'),
		('In the Wake of Poseidon', 'Keith Tippett'),
		('In the Wake of Poseidon', 'Mel Collins'),
		('In the Wake of Poseidon', 'Gordon Haskell'),
		('In the Wake of Poseidon', 'Peter Sinfield')
	) v(AlbumName, MemberName)
) AS src
	ON tgt.AlbumName = src.AlbumName
	AND tgt.MemberName = src.MemberName
WHEN NOT MATCHED BY TARGET THEN INSERT
(
	AlbumName,
	MemberName
)
VALUES
(
	src.AlbumName,
	src.MemberName
);

MERGE [4NF].[MemberXInstrument] AS tgt
USING
(
	SELECT
		MemberName,
		InstrumentName
	FROM
	(
		VALUES
		('Adrian Belew', 'V-Drums'),
		('Bill Bruford', 'Drums'),
		('Gordon Haskell', 'Vocals'),
		('Greg Lake', 'Bass'),
		('Greg Lake', 'Vocals'),
		('Ian McDonald', 'Reeds'),
		('Ian McDonald', 'Woodwind'),
		('Ian McDonald', 'Vibes'),
		('Ian McDonald', 'Keyboards'),
		('Ian McDonald', 'Mellotron'),
		('Ian McDonald', 'Vocals'),
		('Keith Tippett', 'Piano'),
		('Mel Collins', 'Sax'),
		('Mel Collins', 'Flute'),
		('Michael Giles', 'Drums'),
		('Michael Giles', 'Percussion'),
		('Michael Giles', 'Vocals'),
		('Peter Giles', 'Bass'),
		('Peter Sinfield', 'Words'),
		('Peter Sinfield', 'Words and Illumination'),
		('Robert Fripp', 'Guitar'),
		('Robert Fripp', 'Mellotron'),
		('Robert Fripp', 'Devices'),
		('Tony Levin', 'Bass'),
		('Tony Levin', 'Stick'),
		('Tony Levin', 'Synth'),
		('Trey Gunn', 'Guitar synth'),
		('Trey Gunn', 'Touch guitar'),
		('Trey Gunn', 'Talker')
	) v(MemberName, InstrumentName)
) AS src
	ON tgt.MemberName = src.MemberName
	AND tgt.InstrumentName = src.InstrumentName
WHEN NOT MATCHED BY TARGET THEN INSERT
(
	MemberName,
	InstrumentName
)
VALUES
(
	src.MemberName,
	src.InstrumentName
);



--1)  If our hypothesis is correct, we should be able to rebuild
--	the contents of AlbumXMemberXInstrument exactly.  Write a script
--	which does this.
